import 'dart:async';
import 'dart:convert';

import 'package:ampos_flutter_sound/android_encoder.dart';
import 'package:ampos_flutter_sound/ios_quality.dart';
import 'package:ampos_flutter_sound/model/exceptions.dart';
import 'package:ampos_flutter_sound/model/progresses.dart';
import 'package:ampos_flutter_sound/model/status.dart';
import 'package:flutter/services.dart';

class AmposFlutterSound {
  static const MethodChannel _channel = const MethodChannel('com.ampos.fluttersound/sound');
  static StreamController<RecordProgress> _recorderController;
  static StreamController<double> _dbPeakController;
  static StreamController<PlayProgress> _playerController;

  /// Value ranges from 0 to 120
  Stream<double> get onRecorderDbPeakChanged => _dbPeakController?.stream;

  Stream<RecordProgress> get onRecorderStateChanged => _recorderController?.stream;

  Stream<PlayProgress> get onPlayerStateChanged => _playerController?.stream;

  bool get isPlaying => _isPlaying;

  bool get isRecording => _isRecording;

  bool _isRecording = false;
  bool _isPlaying = false;

  final bool passiveRetrieveStatus;

  double _subscriptionDuration = 0.01;

  Timer timer;

  AmposFlutterSound({this.passiveRetrieveStatus = false});

  Future<String> setSubscriptionDuration(double sec) async {
    final String result = await _channel.invokeMethod('setSubscriptionDuration', <String, dynamic>{
      'sec': sec,
    });
    _subscriptionDuration = sec;
    return result;
  }

  Future _setRecorderCallback() async {
    if (_recorderController == null) {
      _recorderController = StreamController.broadcast();
    }
    if (_dbPeakController == null) {
      _dbPeakController = StreamController.broadcast();
    }
    if (passiveRetrieveStatus) {
      timer = getTimer((timer) async {
        final Map<String, dynamic> result = await getRecorderProgress();
        updateRecorderProgress(result);
      });
    } else {
      _channel.setMethodCallHandler((MethodCall call) {
        switch (call.method) {
          case 'updateRecorderProgress':
            final Map<String, dynamic> result = json.decode(call.arguments);
            updateRecorderProgress(result);
            break;
          case 'updateDbPeakProgress':
            if (_dbPeakController != null) _dbPeakController.add(call.arguments);
            break;
          default:
            throw ArgumentError('Unknown method ${call.method} ');
        }
        return null;
      });
    }
  }

  Future _setPlayerCallback() async {
    if (_playerController == null) {
      _playerController = StreamController.broadcast();
    }

    if (passiveRetrieveStatus) {
      timer = getTimer((timer) async {
        final Map<String, dynamic> result = await getPlayerProgress();
        updatePlayerProgress(result);
      });
    } else {
      _channel.setMethodCallHandler((MethodCall call) {
        switch (call.method) {
          case 'updateProgress':
            final Map<String, dynamic> result = jsonDecode(call.arguments);
            updatePlayerProgress(result);
            break;
          case 'audioPlayerDidFinishPlaying':
            final Map<String, dynamic> result = jsonDecode(call.arguments);
            final PlayProgress status = PlayProgress.fromJson(result);
            if (status.currentPosition != status.duration) {
              status.currentPosition = status.duration;
            }
            if (_playerController != null) {
              _playerController.add(status);
            }
            _onPlayerStopped();
            break;
          default:
            throw ArgumentError('Unknown method ${call.method}');
        }
        return null;
      });
    }
  }

  Timer getTimer(void callback(Timer timer)) {
    return Timer.periodic(Duration(milliseconds: (_subscriptionDuration * 1000).floor()), callback);
  }

  Future<Map<String, dynamic>> getRecorderProgress() async {
    return jsonDecode(await _channel.invokeMethod('getRecorderStatus'));
  }

  void updateRecorderProgress(Map<String, dynamic> json) {
    RecordProgress progress = RecordProgress.fromJson(json);
    _recorderController?.add(progress);

    print(json);

    if (passiveRetrieveStatus && progress.status != Status.active) {
      try {
        print(
            'passiveRetrieveStatus $passiveRetrieveStatus, progress.status != Status.active ${progress.status != Status.active}');
        stopRecorder();
      } on RecorderStoppedException catch (e) {}
    }
  }

  Future<Map<String, dynamic>> getPlayerProgress() async {
    return jsonDecode(await _channel.invokeMethod('getPlayerStatus'));
  }

  void updatePlayerProgress(Map<String, dynamic> result) {
    PlayProgress progress = PlayProgress.fromJson(result);
    _playerController?.add(progress);

    if (passiveRetrieveStatus && progress.status != Status.active) {
      stopPlayer();
    }
  }

  void _removeRecorderCallback() {
    if (_recorderController != null) {
      _recorderController
        ..done
        ..close();
      _recorderController = null;
    }
  }

  void _removeDbPeakCallback() {
    if (_dbPeakController != null) {
      _dbPeakController
        ..done
        ..close();
      _dbPeakController = null;
    }
  }

  void _removePlayerCallback() {
    if (_playerController != null) {
      _playerController
        ..done
        ..close();
      _playerController = null;
    }
  }

  Future<String> startRecorder(
      {String uri,
      int sampleRate = 44100,
      int numChannels = 2,
      int bitRate,
      AndroidEncoder androidEncoder = AndroidEncoder.AAC,
      IosQuality iosQuality = IosQuality.HIGH}) async {
    if (this._isRecording) {
      throw RecorderRunningException('Recorder is already recording.');
    }
    try {
      final String result = await _channel.invokeMethod('startRecorder', <String, dynamic>{
        'path': uri,
        'sampleRate': sampleRate,
        'numChannels': numChannels,
        'bitRate': bitRate,
        'androidEncoder': androidEncoder?.value,
        'iosQuality': iosQuality?.value
      });
      _setRecorderCallback();

      _onRecorderStarted();
      return result;
    } catch (err) {
      throw Exception(err);
    }
  }

  void _onRecorderStarted() {
    this._isRecording = true;
  }

  Future<String> stopRecorder() async {
    String result;
    if (this._isRecording) {
      result = await _channel.invokeMethod('stopRecorder');
    }
    _onRecorderStopped();
    return result;
  }

  void _onRecorderStopped() {
    _isRecording = false;
    _removeRecorderCallback();
    _removeDbPeakCallback();
    timer?.cancel();
  }

  Future<String> startPlayer(String uri) async {
    if (_isPlaying) {
      throw PlayerRunningException('Player is already playing.');
    }

    try {
      final String result = await _channel.invokeMethod('startPlayer', <String, dynamic>{
        'path': uri,
      });
      _onPlayerStarted();
      return result;
    } catch (err) {
      throw Exception(err);
    }
  }

  void _onPlayerStarted() {
    _setPlayerCallback();
    _isPlaying = true;
  }

  Future<String> stopPlayer() async {
    String result;
    if (this._isPlaying) {
      result = await _channel.invokeMethod('stopPlayer');
    }

    _onPlayerStopped();
    return result;
  }

  void _onPlayerStopped() {
    _isPlaying = false;
    _removePlayerCallback();
    timer?.cancel();
  }

  Future<String> pausePlayer() async {
    final String result = await _channel.invokeMethod('pausePlayer');
    return result;
  }

  Future<String> resumePlayer() async {
    final String result = await _channel.invokeMethod('resumePlayer');
    return result;
  }

  Future<String> seekToPlayer(int milliSecs) async {
    final String result = await _channel.invokeMethod('seekToPlayer', <String, dynamic>{
      'sec': milliSecs,
    });
    return result;
  }

  Future<String> setVolume(double volume) async {
    String result = '';
    if (volume < 0.0 || volume > 1.0) {
      result = 'Value of volume should be between 0.0 and 1.0.';
      return result;
    }

    result = await _channel.invokeMethod('setVolume', <String, dynamic>{
      'volume': volume,
    });
    return result;
  }

  /// Defines the interval at which the peak level should be updated.
  /// Default is 0.8 seconds
  Future<String> setDbPeakLevelUpdate(double intervalInSecs) async {
    final String result = await _channel.invokeMethod('setDbPeakLevelUpdate', <String, dynamic>{
      'intervalInSecs': intervalInSecs,
    });
    return result;
  }

  /// Enables or disables processing the Peak level in db's. Default is disabled
  Future<String> setDbLevelEnabled(bool enabled) async {
    final String result = await _channel.invokeMethod('setDbLevelEnabled', <String, dynamic>{
      'enabled': enabled,
    });
    return result;
  }
}
